import org.apache.spark.ml.{Pipeline, PipelineModel}
import org.apache.spark.ml.linalg.Vector
import org.apache.spark.sql.{Column, Row}
import com.datastax.spark.connector._
import org.apache.spark.SparkContext
import org.apache.spark.SparkContext._
import org.apache.spark.SparkConf
import org.apache.spark.sql.DataFrame
import org.apache.spark.ml.classification._
import scala.math.sqrt
import org.apache.spark.ml.linalg.Vectors
import org.apache.spark.ml.linalg.Vectors._
import org.apache.spark.ml.feature._
import org.apache.spark.ml.tuning.ParamGridBuilder
import org.apache.spark.ml.tuning.CrossValidator
import org.apache.spark.ml.evaluation._

import org.apache.spark.sql.SQLContext


val sqlContext = new SQLContext(sc);
import sqlContext.implicits._;

val dsHealthRelatedTweetTexts = sqlContext.read.textFile("data/HealthRelatedTweetTexts_12-8-2016.txt")
val dsNonHealthRelatedTweetTexts = sqlContext.read.textFile("data/NonHealthRelatedTweetTexts_12-8-2016.txt")

val dsLabeledHealthRelatedTweetTexts = dsHealthRelatedTweetTexts.withColumn("label", lit(1.0))
val dsLabeledNonHealthRelatedTweetTexts = dsNonHealthRelatedTweetTexts.withColumn("label", lit(0.0))

val dsAllTweetTexts = dsLabeledHealthRelatedTweetTexts.union(dsLabeledNonHealthRelatedTweetTexts)
val fractionSubset1 = 0.5
val Array(dfLabeledSubset1, dfLabeledSubset2) = dsAllTweetTexts.randomSplit(Array(fractionSubset1, 1-fractionSubset1))


val wordSplitter = new Tokenizer().setInputCol("value").setOutputCol("tokens")
wordSplitter.transform(dfLabeledSubset1).show

val trivialWordRemover = new StopWordsRemover().setInputCol("tokens").setOutputCol("significantWords")
// show the significant words
trivialWordRemover.transform(wordSplitter.transform(dsAllTweetTexts)).show

val mapWordsToPresenceAbsenceVector = new CountVectorizer().setInputCol("significantWords").setOutputCol("features").setVocabSize(65536).setMinDF(3)

val logisticRegression = new LogisticRegression().setMaxIter(100000) // .setRegParam(0.2).setMaxIter(100000).setTol(0.001).setStandardization(true)

val logisticRegressionPipeline = new Pipeline().setStages(Array(wordSplitter, trivialWordRemover, mapWordsToPresenceAbsenceVector, logisticRegression))

val logisticRegressionModelWithPipeline = logisticRegressionPipeline.fit(dsAllTweetTexts)
// having been fit applying the model to a dataset applies the ENTIRE PIPELINE to that dataset!
val dfLogisticRegressionFromPipeline = logisticRegressionModelWithPipeline.transform(dsAllTweetTexts)

val binaryClassificationEvaluator = new BinaryClassificationEvaluator().setMetricName("areaUnderROC").setRawPredictionCol("rawPrediction").setLabelCol("label")

binaryClassificationEvaluator.evaluate(dfLogisticRegressionFromPipeline)


// Ok, now perform manual cross-validation
// phase 1
val logisticRegressionModelWithPipelineTrainedOn1 = logisticRegressionPipeline.fit(dfLabeledSubset1)
val dfLogisticRegressionFromPipelineTrainedOn1 = logisticRegressionModelWithPipelineTrainedOn1.transform(dfLabeledSubset2)
binaryClassificationEvaluator.evaluate(dfLogisticRegressionFromPipelineTrainedOn1)

// phase 2
val logisticRegressionModelWithPipelineTrainedOn2 = logisticRegressionPipeline.fit(dfLabeledSubset2)
val dfLogisticRegressionFromPipelineTrainedOn2 = logisticRegressionModelWithPipelineTrainedOn2.transform(dfLabeledSubset1)

binaryClassificationEvaluator.evaluate(dfLogisticRegressionFromPipelineTrainedOn2)


// now examine the information on the structure of the model -- which words are most helpful in the classification
// examine the stages
logisticRegressionModelWithPipeline.stages.length
// examine the logistic regression
logisticRegressionModelWithPipeline.stages(3).extractParamMap
val bestLRModel = logisticRegressionModelWithPipeline.stages(3).asInstanceOf[org.apache.spark.ml.classification.LogisticRegressionModel]

// examine the coefficients
bestLRModel.coefficients
val coeffs = bestLRModel.coefficients
coeffs.numActives
// get the index of associated with the largest coefficient (indicating importance in the logistic regression model)
coeffs.argmax

// now extract the word to vector model
val wordToVectorMapperModel = logisticRegressionModelWithPipeline.stages(2).asInstanceOf[org.apache.spark.ml.feature.CountVectorizerModel]
wordToVectorMapperModel.vocabulary

wordToVectorMapperModel.vocabulary(coeffs.argmax)

// find out words that are highly predictive of health classification
coeffs.toArray.zipWithIndex.collect { case (v, i) if (v > 40) => wordToVectorMapperModel.vocabulary(i) }
coeffs.toArray.zipWithIndex.collect { case (v, i) if (v > 100) => wordToVectorMapperModel.vocabulary(i) }
coeffs.toArray.zipWithIndex.collect { case (v, i) if (v > 140) => wordToVectorMapperModel.vocabulary(i) }
// words highly predictive of NON health
coeffs.toArray.zipWithIndex.collect { case (v, i) if (v < -140) => wordToVectorMapperModel.vocabulary(i) }

/// Now on to layering this on a cross validation pipeline


val paramGridForLogisticRegressionCrossValidation = new ParamGridBuilder().addGrid(logisticRegression.regParam, Array(0.05, 0.20, 0.40, 0.6, 0.8)).addGrid(logisticRegression.tol, Array(0.001, 0.01, 0.1)).build()

val crossValidatorForLogisticRegression = new CrossValidator().setEstimator(logisticRegressionPipeline).setEvaluator(binaryClassificationEvaluator).setEstimatorParamMaps(paramGridForLogisticRegressionCrossValidation).setNumFolds(10)

val crossValidationResultsForLogisticRegression = crossValidatorForLogisticRegression.fit(dsAllTweetTexts)
crossValidationResultsForLogisticRegression.avgMetrics
val logisticRegressionModelFromCrossValidation = crossValidationResultsForLogisticRegression.bestModel.asInstanceOf[org.apache.spark.ml.PipelineModel]
logisticRegressionModelFromCrossValidation.parent.extractParamMap

val dfPredictedFromCrossValidatedLogisticRegressionModel = logisticRegressionModelFromCrossValidation.transform(dsAllTweetTexts)
dfPredictedFromCrossValidatedLogisticRegressionModel.select("probability", "prediction", "label").show(100)

binaryClassificationEvaluator.evaluate(dfPredictedFromCrossValidatedLogisticRegressionModel)

logisticRegressionModelFromCrossValidation.write.overwrite().save("output/HealthRecognitionModel_Dec-10-2016")
